(*
  tali.v

  A version of TAL with polymorphism over integers, implemented
  using HOAS in Coq

  Nadeem Abdul Hamid, February 2006
*)

Require Import Compare_dec.
Require Import Eqdep.
Require Import EqNat.
Require Import List.
Require Import natutil.
Require Import tactics.


Open Scope type_scope.

(* TAL_i SYNTAX *)

(* TAL_i registers *)
Inductive reg : Set := r0 | r1 | r2 | r3 | r4 | r5 | r6 | r7.

(* TAL_i types *)
Inductive ttype : Set :=
  | ttop : ttype
  | tint : nat -> ttype
  | tcode : ctype -> ttype
with ctype : Set :=
  | cbase : (reg -> ttype) -> ctype
  | cpnat : (nat -> ctype) -> ctype
.

(* register file type *)
Definition rftype := reg -> ttype.

(* Code addresses *)
Definition addr := nat.

(* List of integers *)
Definition natlist := list nat.

(* word values *)
Inductive wval : Set :=
  | wint : nat -> wval
  | wcaddr : addr -> wval
  | wappi : wval -> nat -> wval
.

(* register file *)
Definition regfile := reg -> wval.

(* instructions *)
Inductive instr : Set :=
  | iadd : reg -> reg -> reg -> instr
  | iaddi : reg -> reg -> nat -> instr
  | isub  : reg -> reg -> reg -> instr
  | isubi : reg -> reg -> nat -> instr
  | imov  : reg -> reg -> instr
  | imovi : reg -> nat -> instr
  | imovf : reg -> addr -> instr
  | ibgti : reg -> nat -> addr -> natlist -> instr
  | itappi : reg -> nat -> instr
.

(* ibgti and ijd need the natlist to tie things together for the
   preservation lemmas - when doing inversion on evalxm and wf_iseq *)
(* for ijmp we require that the value in the register has
   been fully applied... *)
Inductive iseq : Set :=
  | icons : instr -> iseq -> iseq
  | ijd : addr -> natlist -> iseq
  | ijmp : reg -> iseq
.

(* Code instruction sequences under some variable
   environment *)
Fixpoint ciseq (t:ctype) : Set := 
  match t with
    | cbase _ => iseq
    | cpnat F => forall (i:nat), ciseq (F i)
  end.

(* code values are strong pairs of code type and code instruction sequence *)
Definition codeval := {t:ctype & ciseq t}. (* sigS (fun i => ciseq i).*)

(* Code heap *)
Definition codeheap := addr -> option codeval.

(* Program state *)
Definition progstate := (codeheap * regfile * iseq).



(************************************************************************)
(* Type application definitions *)

Inductive app_ctype : ctype -> natlist -> rftype -> Prop :=
  | app_cbase : forall G, app_ctype (cbase G) nil G
  | app_cpnat : forall F n ns G,
                app_ctype (F n) ns G ->
		app_ctype (cpnat F) (n::ns) G
.

Inductive app_ciseq : forall t, ciseq t -> natlist -> iseq -> Prop :=
  | app_cibase : forall G Is, app_ciseq (cbase G) Is nil Is
  | app_cipnat : forall F cis n ns Is,
                 app_ciseq (F n) (cis n) ns Is ->
		 app_ciseq (cpnat F) cis (n::ns) Is
.

Fixpoint snoc (A:Set) (a:A) (As:list A) {struct As}: list A :=
  match As with
    | nil => a :: nil
    | a' :: As' => a' :: (snoc A a As')
  end.

Fixpoint terase (v:wval) : wval :=
  match v with
    | wappi v _ => terase v
    | v => v
  end.

Fixpoint tcollect (v:wval) : natlist :=
  match v with
    | wappi v n => snoc _ n (tcollect v)
    | _ => nil
  end.




(************************************************************************)
(* TAL_i Operational Semantics *)

(* Some utility functions *)
Definition beq_reg_sumbool : forall (r:reg) r', {r=r'}+{r<>r'}.
Proof.
  double induction r r'; intros; auto; try (right; discriminate).
Defined.

Definition beq_reg : forall (r r':reg), bool.
Proof.
  intros; elim (beq_reg_sumbool r r'); intros; [ exact true | exact false ].
Defined.

Definition rfupd (R:regfile) (r:reg) (v:wval) : regfile
  := fun r' => if (beq_reg r' r) then v else R r'.


Inductive evalp : progstate -> progstate -> Prop :=
  | ev_iadd : forall C R Is rd rs rt s t,
    R rs = wint s ->
    R rt = wint t ->      
    evalp (C, R, icons (iadd rd rs rt) Is) 
         (C, (rfupd R rd (wint (s+t))), Is)
  | ev_iaddi : forall C R Is rd rs i s,
    R rs = wint s ->
    evalp (C, R, icons (iaddi rd rs i) Is) 
         (C, (rfupd R rd (wint (s+i))), Is)
  | ev_isub : forall C R Is rd rs rt s t,
    R rs = wint s ->
    R rt = wint t ->      
    evalp (C, R, icons (isub rd rs rt) Is) 
         (C, (rfupd R rd (wint (s-t))), Is)
  | ev_isubi : forall C R Is rd rs i s,
    R rs = wint s ->
    evalp (C, R, icons (isubi rd rs i) Is) 
         (C, (rfupd R rd (wint (s-i))), Is)
  | ev_imov : forall C R Is rd rs,
    evalp (C, R, icons (imov rd rs) Is) 
         (C, (rfupd R rd (R rs)), Is)
  | ev_imovi : forall C R Is rd i,
    evalp (C, R, icons (imovi rd i) Is) 
         (C, (rfupd R rd (wint i)), Is)
  | ev_imovf : forall C R Is rd f,
    evalp (C, R, icons (imovf rd f) Is) 
         (C, (rfupd R rd (wcaddr f)), Is)
  | ev_itappi : forall C R Is rd i,
    evalp (C, R, icons (itappi rd i) Is) 
         (C, (rfupd R rd (wappi (R rd) i)), Is)
  | ev_ibgtiT : forall (C:codeheap) R Is rs i f ns s t cis Is',
    R rs = wint s ->
    s > i ->
    C f = Some (existS _ t cis) ->
    app_ciseq t cis ns Is' ->
    evalp (C, R, icons (ibgti rs i f ns) Is) (C, R, Is')
  | ev_ibgtiF : forall C R Is rs i f ns s,
    R rs = wint s ->
    s <= i ->
    evalp (C, R, icons (ibgti rs i f ns) Is) (C, R, Is)
  | ev_ijd : forall (C:codeheap) R f ns t cis Is,
    C f = Some (existS _ t cis) ->
    app_ciseq t cis ns Is ->
    evalp (C, R, ijd f ns) (C, R, Is)
  | ev_ijmp : forall (C:codeheap) R r f ns t cis Is,
    terase (R r) = wcaddr f ->
    tcollect (R r) = ns ->
    C f = Some (existS _ t cis) ->
    app_ciseq t cis ns Is ->
    evalp (C, R, ijmp r) (C, R, Is)
.



(************************************************************************)
(* TAL_i Static Semantics *)

(* Register file subtyping *)
Definition subrftype (G1 G2:rftype) : Prop :=
  forall r, (G1 r) = (G2 r) \/ (G2 r)=ttop.

Definition ctypeOf (c:codeval) : ctype := (projS1 c).

Inductive wf_wval : codeheap -> wval -> ttype -> Prop :=
  | wf_wtop : forall C v, wf_wval C v ttop
  | wf_wint : forall C i, wf_wval C (wint i) (tint i)
  | wf_wcaddr : forall C f cv, 
    C f = Some cv -> wf_wval C (wcaddr f) (tcode (ctypeOf cv))
  | wf_wappi : forall C v n F,
    wf_wval C v (tcode (cpnat F)) -> wf_wval C (wappi v n) (tcode (F n))
.

Inductive wf_regfile : codeheap -> regfile -> rftype -> Prop :=
  | wf_rfile : forall C R G,
               (forall r, wf_wval C (R r) (G r)) -> wf_regfile C R G
.

Definition rftupd (G:rftype) (r:reg) (t:ttype) : rftype
  := fun r' => if (beq_reg r' r) then t else G r'.

Inductive wf_iseq : codeheap -> rftype -> iseq -> Prop :=
  | wf_iadd : forall C G rd rs rt Is i j,
    G rs = tint i -> G rt = tint j ->
    wf_iseq C (rftupd G rd (tint (i+j))) Is ->
    wf_iseq C G (icons (iadd rd rs rt) Is)
  | wf_iaddi : forall C G rd rs t Is i,
    G rs = tint i -> 
    wf_iseq C (rftupd G rd (tint (i+t))) Is ->
    wf_iseq C G (icons (iaddi rd rs t) Is)
  | wf_isub : forall C G rd rs rt Is i j,
    G rs = tint i -> G rt = tint j ->
    wf_iseq C (rftupd G rd (tint (i-j))) Is ->
    wf_iseq C G (icons (isub rd rs rt) Is)
  | wf_isubi : forall C G rd rs t Is i,
    G rs = tint i -> 
    wf_iseq C (rftupd G rd (tint (i-t))) Is ->
    wf_iseq C G (icons (isubi rd rs t) Is)
  | wf_imov : forall C G rd rs Is,
    wf_iseq C (rftupd G rd (G rs)) Is ->
    wf_iseq C G (icons (imov rd rs) Is)
  | wf_imovi : forall C G rd i Is,
    wf_iseq C (rftupd G rd (tint i)) Is ->
    wf_iseq C G (icons (imovi rd i) Is)
  | wf_imovf : forall C G rd f Is cv,
    C f = Some cv ->
    wf_iseq C (rftupd G rd (tcode (ctypeOf cv))) Is ->
    wf_iseq C G (icons (imovf rd f) Is)
  | wf_itappi : forall C G rd n Is F,
    G rd = tcode (cpnat F) ->
    wf_iseq C (rftupd G rd (tcode (F n))) Is ->
    wf_iseq C G (icons (itappi rd n) Is)
  | wf_ibgti : forall C G rs i f ns Is s cv G',
    G rs = tint s ->
  (* for the true case *)
    C f = Some cv ->
    app_ctype (ctypeOf cv) ns G' ->
    subrftype G G' ->
  (* for the false case *)
    wf_iseq C G Is ->
    wf_iseq C G (icons (ibgti rs i f ns) Is)
  | wf_ijd : forall C G f ns cv G',
    C f = Some cv ->
    app_ctype (ctypeOf cv) ns G' ->
    subrftype G G' ->
    wf_iseq C G (ijd f ns)
  | wf_ijmp : forall C G r G',
    G r = tcode (cbase G') ->
    subrftype G G' ->
    wf_iseq C G (ijmp r)
.

Inductive wf_ciseq : codeheap -> forall t, ciseq t -> Prop :=
  | wf_cibase : forall C G Is, wf_iseq C G Is -> wf_ciseq C (cbase G) Is
  | wf_cipnat : forall C t cis,
    (forall n, wf_ciseq C (t n) (cis n)) ->  wf_ciseq C (cpnat t) cis
.

Definition wf_codeval : codeheap -> codeval -> Prop := 
  fun C cv => match cv with existS t cis => wf_ciseq C t cis end.

Inductive wf_codeheap : codeheap -> Prop :=
  | wf_cheap : forall C, 
    (forall f cv, C f = Some cv -> wf_codeval C cv) ->
    wf_codeheap C
.

Inductive wf_progstate : progstate -> Prop :=
  | wf_ps : forall C R Is G,
    wf_codeheap C ->
    wf_regfile C R G ->
    wf_iseq C G Is ->
    wf_progstate (C, R, Is).



(************************************************************************)
(* TAL_i SAFETY PROOFS *)

(* If code type can be instantiated to base, then so can the 
   corresponding code value (used by progress of jump instructions) *)
Lemma app_ctype_ciseq
  : forall t ns G,
    app_ctype t ns G -> forall cis, exists Is, app_ciseq t cis ns Is.
Proof.
  intros t ns G H; induction H; intros.
  exists cis; constructor.

  mcunfoldFoldIn ciseq (ciseq (cpnat F)).
  elim (IHapp_ctype (cis n)); intros; exists x.
  constructor; auto.
Qed.

Lemma app1_snoc : forall A (cs:list A) (c:A), cs ++ c :: nil = snoc _ c cs.
Proof.
  induction cs; unfoldFold snoc; intros; auto.
  replace ((a::cs)++c ::nil) with (a:: (cs++c :: nil)); auto.
  rewrite (IHcs c); auto.
Qed.

Lemma app_nil_eq : forall A (cs:list A), cs++nil = cs.
Proof.
  induction cs; compute; auto.
  unfold app in IHcs; rewrite IHcs; auto.
Qed.

Hint Immediate app1_snoc app_nil_eq.


(* Canoncial forms lemmas *)

Lemma canform_rf_int
  : forall C R G r s,
    wf_regfile C R G ->
    G r = tint s ->
    R r = wint s.
Proof.
  intros; mcinvClear wf_regfile.
  generalize (H1 r); mcrewrite (G r); intros; mcinvClear wf_wval.
Qed.


Lemma canform_wf_tcode_aux
  : forall C v t,
    wf_wval C v t ->
    forall ct ns G, t = tcode ct ->
    app_ctype ct ns G ->
    exists f, exists cv, exists ns', exists ns'',
      terase v = wcaddr f /\
      tcollect v = ns' /\
      ns'' = ns' ++ ns /\
      C f = Some cv /\
      app_ctype (ctypeOf cv) ns'' G.
Proof.
  intros C v t H; induction H; intros; try discriminate. 
  existsiv f cv (@nil nat) ns.
  injection H0; intros; mcrewriteRev ct.

  assert (app_ctype (cpnat F) (n::ns) G).
    injection H0; constructor; try mcrewriteRev ct.
  generalize (IHwf_wval (cpnat F) (n::ns) G (refl_equal _) H2);
    intros (f, (cv, (ns', (ns'', HH)))); splitAllHyps.
  existsiv f cv (ns' ++ n::nil) ns''.
  concsplit; auto.
  unfoldFold tcollect; mcrewrite (tcollect v); symmetry; apply (app1_snoc nat).
  replace ((ns'++n::nil)++ns) with (ns' ++ (n::nil ++ ns)); auto.
  symmetry; apply (app_ass ns' (n::nil) ns).
Qed.


Lemma canform_wf_tcode
  : forall C v G,
    wf_wval C v (tcode (cbase G)) ->
    exists f, exists cv, exists ns,
      terase v = wcaddr f /\
      tcollect v = ns /\
      C f = Some cv /\
      app_ctype (ctypeOf cv) ns G.
Proof.
  intros.
  generalize
    (canform_wf_tcode_aux C v (tcode (cbase G)) H (cbase G) nil G (refl_equal _)
      (app_cbase _)).
  intros (f, (cv, (ns', (ns'', HH)))); splitAllHyps.
  existsiii f cv ns''; concsplit; auto.
  mcrewrite ns''. 
  replace (ns'++nil) with ns'; auto.
  symmetry; apply (app_nil_eq nat).
Qed.


Lemma wf_wval_tcode_app_ctype
  : forall C v G f ns cv,
    wf_wval C v (tcode (cbase G)) ->
    terase v = wcaddr f ->
    tcollect v = ns ->
    C f = Some cv ->
    app_ctype (ctypeOf cv) ns G.
Proof.
  intros.
  generalize (canform_wf_tcode _ _ _ H).
  intros (f', (cv', (ns', HH))); splitAllHyps.
  mcinjection (terase v); mcinjection (tcollect v).
  mcrewriteIn f' (C f'=Some cv'); mcinjection (C f).
  mcrewrite2 cv ns.
Qed.


(* Progress Lemmas *)

Ltac progressStartTac curIs :=
  intros; unfold curIs; mcunfoldIn curIs wf_progstate; clear curIs;
    mcinvClear wf_progstate; mcinvClear wf_iseq.


Lemma progress_iadd
  : forall C R rd rs rt Is,
    let curIs := (icons (iadd rd rs rt) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  generalize (canform_rf_int _ _ _ _ _ H1 H); intros.
  generalize (canform_rf_int _ _ _ _ _ H1 H3); intros.
  exists (C, (rfupd R rd (wint (i+j))), Is);
    constructor; auto.
Qed.


Lemma progress_iaddi
  : forall C R rd rs t Is,
    let curIs := (icons (iaddi rd rs t) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  generalize (canform_rf_int _ _ _ _ _ H1 H); intros.
  exists (C, (rfupd R rd (wint (i+t))), Is);
    constructor; auto.
Qed.


Lemma progress_isub
  : forall C R rd rs rt Is,
    let curIs := (icons (isub rd rs rt) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  generalize (canform_rf_int _ _ _ _ _ H1 H); intros.
  generalize (canform_rf_int _ _ _ _ _ H1 H3); intros.
  exists (C, (rfupd R rd (wint (i-j))), Is);
    constructor; auto.
Qed.


Lemma progress_isubi
  : forall C R rd rs t Is,
    let curIs := (icons (isubi rd rs t) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  generalize (canform_rf_int _ _ _ _ _ H1 H); intros.
  exists (C, (rfupd R rd (wint (i-t))), Is);
    constructor; auto.
Qed.


Lemma progress_imov
  : forall C R rd rs Is,
    let curIs := (icons (imov rd rs) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  exists (C, (rfupd R rd (R rs)), Is);
    constructor; auto.
Qed.


Lemma progress_imovi
  : forall C R rd s Is,
    let curIs := (icons (imovi rd s) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  exists (C, (rfupd R rd (wint s)), Is);
    constructor; auto.
Qed.


Lemma progress_imovf
  : forall C R rd f Is,
    let curIs := (icons (imovf rd f) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  exists (C, (rfupd R rd (wcaddr f)), Is);
    constructor; auto.
Qed.


Lemma progress_itappi
  : forall C R rd n Is,
    let curIs := (icons (itappi rd n) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  exists (C, (rfupd R rd (wappi (R rd) n)), Is);
    constructor; auto.
Qed.


Lemma progress_ibgti
  : forall C R rs i f ns Is,
    let curIs := (icons (ibgti rs i f ns) Is) in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  generalize (canform_rf_int _ _ _ _ _ H1 H); intros.
  
  elim (le_gt_dec s i); intros; auto.
  exists (C, R, Is); eapply ev_ibgtiF; eauto.

  generalize cv H3 H4; clear cv H3 H4; intros (t, cis); unfold ctypeOf; intros.
  elim (app_ctype_ciseq _ _ _ H4 cis); unfold projS1; intros Is' His'.
  exists (C, R, Is'); apply ev_ibgtiT with s t cis; auto.
Qed.


Lemma progress_ijd
  : forall C R f ns,
    let curIs := ijd f ns in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  generalize cv H H3; clear cv H H3.
  intros (t, cis); unfold ctypeOf; intros.
  elim (app_ctype_ciseq _ _ _ H3 cis); unfold projS1; intros Is' His'.
  exists (C, R, Is'); apply ev_ijd with t cis; auto.
Qed.


Lemma progress_ijmp
  : forall C R r,
    let curIs := ijmp r in
      wf_progstate (C, R, curIs) -> exists P, evalp (C, R, curIs) P.
Proof.
  progressStartTac curIs.
  assert (wf_wval C (R r) (tcode (cbase G'))).
    mcinvClear wf_regfile.
    mcrewriteRev (G r); eapplyinf.
  generalize (canform_wf_tcode _ _ _ H2); 
    intros (f, ((t,cis), (ns, HH))); splitAllHyps.
  elim (app_ctype_ciseq _ _ _ H8 cis); unfold ctypeOf, projS1; intros Is' His'.
  exists (C, R, Is'); apply ev_ijmp with f ns t cis; auto.
Qed.

  

(* Utility lemmas for preservation proofs *)

Lemma beq_reg_true
  : forall r, true = beq_reg r r.
Proof.
  intros; unfold beq_reg; assert (r=r) ;auto.
  elim (beq_reg_sumbool r r); intros; auto; try absurdEq.
Qed.

Lemma beq_reg_false1
  : forall r r', r<>r' -> false = beq_reg r r'.
Proof.
  intros; unfold beq_reg; elim (beq_reg_sumbool r r'); intros; auto;
    try absurdEq.
Qed.

Lemma beq_reg_false2
  : forall r r', r'<>r -> false = beq_reg r r'.
Proof.
  intros; unfold beq_reg; elim (beq_reg_sumbool r r'); intros; auto;
    try absurdEq.
Qed.

Hint Immediate beq_reg_true.
Hint Resolve beq_reg_false1 beq_reg_false2.

Lemma subrftype_rftupd
  : forall G G' r t,
    subrftype G G' ->
    subrftype (rftupd G r t) (rftupd G' r t).
Proof.
  unfold subrftype; intros G G' r t H r'.
  unfold rftupd; elim (beq_reg r' r); auto.
Qed.

Hint Resolve subrftype_rftupd.

Lemma subrftype_trans
  : forall G G' G'',
    subrftype G G' ->
    subrftype G' G'' ->
    subrftype G G''.
Proof.
  unfold subrftype; intros.
  generalize (H r) (H0 r); intros.
  elim H1; elim H2; intros; auto.
  mcrewrite (G r).
  mcrewrite (G'' r).
Qed.

Lemma wf_iseq_subrftype
  : forall C G Is, 
    wf_iseq C G Is -> forall G', subrftype G' G -> wf_iseq C G' Is.
Proof.
  intros C G Is H; induction H; intros;
  repeat match goal with
	   H:G ?r=_, H1:subrftype _ G |- _ => 
	     elim (H1 r); rewrite H; clear H; intros; try discriminate
	 end.

  apply wf_iadd with i j; auto;
    try mcrewrite (G' rs); try mcrewrite (G' rt); apply IHwf_iseq; auto.
  apply wf_iaddi with i; auto;
    try mcrewrite (G' rs);  apply IHwf_iseq; auto.
  apply wf_isub with i j; auto;
    try mcrewrite (G' rs); try mcrewrite (G' rt); apply IHwf_iseq; auto.
  apply wf_isubi with i; auto;
    try mcrewrite (G' rs);  apply IHwf_iseq; auto.
  constructor; auto.
    apply IHwf_iseq; auto.
    elim (H0 rs); intros; try mcrewrite (G' rs).
    mcrewrite (G rs); intro;
      unfold rftupd; elim (beq_reg r rd); auto.
  constructor; auto.
  apply wf_imovf with cv; auto.
  apply wf_itappi with F; auto.
  apply wf_ibgti with s cv G'; auto.
    apply subrftype_trans with G; auto.
  apply wf_ijd with cv G'; auto.
    apply subrftype_trans with G; auto.
  apply wf_ijmp with G'; auto.
    apply subrftype_trans with G; auto.
Qed.

(* this lemma needed because coq has trouble with unification
   in you just try inversion tactic *)
Lemma wf_cipnat_inv
  : forall C t cis,
    wf_ciseq C t cis ->
    forall F cis',
      t = cpnat F ->
      (existS _ t cis)=(existS _ (cpnat F) cis') ->
      (forall n, wf_ciseq C (F n) (cis' n)).
Proof.
  intros C t cis H; inversion H; clear H.

  intros. injection H3; intros; rewrite H in H5; discriminate.

  intros F cis' H4; generalize cis'; clear cis'. 
  rewrite H4 in H2; generalize H4; inversion H2; mcrewrite F.
  clear F H4 H2 H5.
  intros e; generalize cis H3; clear cis H3; rewrite e; mcclear t.
  intros; repeat existSelim.
  mcrewriteRev cis; mcrewriteRev cis0.
Qed.


Lemma wf_ciseq_app_wf_iseq
  : forall C t cis ns Is,
    app_ciseq t cis ns Is ->
    forall G,
      app_ctype t ns G ->
      wf_ciseq C t cis ->
      wf_iseq C G Is.
Proof.
  intros C t cis ns Is H; induction H; intros.
  mcinv wf_ciseq; existSelim; mcrewrite Is.
  mcinv app_ctype; mcrewrite G0.

  mcinvClear app_ctype.
  apply IHapp_ciseq; auto.
  apply wf_cipnat_inv with (cpnat F) (cis); auto.
Qed.




(* Preservation Lemmas *)

Definition preservStatement C R i Is P'
  := let curIs := icons i Is in
     wf_progstate (C, R, curIs) -> evalp (C, R, curIs) P' ->
     wf_progstate P'.

Lemma preserv_iadd
  : forall C R rd rs rt Is P',
    preservStatement C R (iadd rd rs rt) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd (tint (s+t))); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
  generalize (canform_rf_int _ _ _ _ _ H4 H); mcrewrite (R rs); intros e;
    injection e; clear e; intros; mcrewrite s.
  generalize (canform_rf_int _ _ _ _ _ H4 H0); mcrewrite (R rt); intros e;
    injection e; clear e; intros; mcrewrite t.
Qed.


Lemma preserv_iaddi
  : forall C R rd rs i Is P',
    preservStatement C R (iaddi rd rs i) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd (tint (s+i))); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
  generalize (canform_rf_int _ _ _ _ _ H3 H); mcrewrite (R rs); intros e;
    injection e; clear e; intros; mcrewrite s.
Qed.


Lemma preserv_isub
  : forall C R rd rs rt Is P',
    preservStatement C R (isub rd rs rt) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd (tint (s-t))); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
  generalize (canform_rf_int _ _ _ _ _ H4 H); mcrewrite (R rs); intros e;
    injection e; clear e; intros; mcrewrite s.
  generalize (canform_rf_int _ _ _ _ _ H4 H0); mcrewrite (R rt); intros e;
    injection e; clear e; intros; mcrewrite t.
Qed.


Lemma preserv_isubi
  : forall C R rd rs i Is P',
    preservStatement C R (isubi rd rs i) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd (tint (s-i))); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
  generalize (canform_rf_int _ _ _ _ _ H3 H); mcrewrite (R rs); intros e;
    injection e; clear e; intros; mcrewrite s.
Qed.


Lemma preserv_imov
  : forall C R rd rs Is P',
    preservStatement C R (imov rd rs) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd (G rs)); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; mcinvClear wf_regfile.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
Qed.


Lemma preserv_imovi
  : forall C R rd i Is P',
    preservStatement C R (imovi rd i) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  apply wf_ps with (rftupd G rd (tint i)); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
  mcinvClear wf_iseq.
Qed.


Lemma preserv_imovf
  : forall C R rd f Is P',
    preservStatement C R (imovf rd f) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  mcinvClear wf_iseq.
  apply wf_ps with (rftupd G rd (tcode (ctypeOf cv))); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor; auto.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
Qed.


Lemma preserv_itappi
  : forall C R rd n Is P',
    preservStatement C R (itappi rd n) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros;
    assert (C=C' /\ Is'=Is); [ mcinv evalp | mcinvClear evalp ]; splitAllHyps;
    mcrewrite Is'; mcclear Is'; mcinvClear wf_progstate.
    mcrewrite C'; mcclear C'.
  mcinvClear wf_iseq.
  apply wf_ps with (rftupd G rd (tcode (F n))); intros; auto;
    try constructor; intros.
  compare r rd; intros e; try rewrite e; unfold rfupd, rftupd.
  replace (beq_reg rd rd) with true; auto; constructor; auto.
    mcrewriteRev (G rd); mcinvClear wf_regfile.
  replace (beq_reg r rd) with false; auto; mcinvClear wf_regfile.
Qed.


Lemma preserv_ibgti
  : forall C R rs i f ns Is P',
    preservStatement C R (ibgti rs i f ns) Is P'.
Proof.
  intros; generalize P'; clear P'; intros ((C', R'), Is').
  unfold preservStatement; intros.
  assert (C'=C).
    mcinv evalp.
  mcrewrite C'; mcrewriteIn C' evalp; mcclear C'.

  mcinv evalp;

  mcclear5 C0 R0 rs0 i0 f0; mcclear3 ns0 Is0 Is'0;
  mcinvClear wf_progstate; mcinvClear wf_iseq;
  apply wf_ps with G; auto;
  try mcrewrite R'; try mcrewrite Is'.
  
  assert (wf_codeval C cv).
     mcinv wf_codeheap; eapplyinf.
  mcinjection (C f).
  mcrewriteIn cv wf_codeval; mcunfoldInSelf wf_codeval.
  assert (wf_iseq C G' Is').
    mcrewriteIn cv app_ctype;
    apply (wf_ciseq_app_wf_iseq _ _ _ _ _ H13 _ H6 H3).
  apply wf_iseq_subrftype with G'; auto.
Qed.


Lemma preserv_ijd
  : forall C R f ns P',
    let curIs := (ijd f ns) in
      wf_progstate (C, R, curIs) -> evalp (C, R, curIs) P' ->
      wf_progstate P'.
Proof.
  intros C R f ns ((C', R'), Is'); intros.
  assert (C'=C).
    mcinv evalp.
  mcrewrite C'; mcrewriteIn C' evalp; mcclear C'.

  mcinv evalp;

  mcclear5 C0 R0 rs0 i0 f0; mcclear3 ns0 Is0 Is'0;
  mcinvClear wf_progstate; mcinvClear wf_iseq;
  apply wf_ps with G; auto;
  try mcrewrite R'; try mcrewrite Is'.
  
  assert (wf_codeval C cv).
     mcinv wf_codeheap; eapplyinf.
  mcinjection (C f).
  mcrewriteIn cv wf_codeval; mcunfoldInSelf wf_codeval.
  assert (wf_iseq C G' Is').
    mcrewriteIn cv app_ctype;
    apply (wf_ciseq_app_wf_iseq _ _ _ _ _ H8 _ H5 H4).
  apply wf_iseq_subrftype with G'; auto; try mcrewrite Is.
Qed.


Lemma preserv_ijmp
  : forall C R r P',
    let curIs := (ijmp r) in
      wf_progstate (C, R, curIs) -> evalp (C, R, curIs) P' ->
      wf_progstate P'.
Proof.
  intros C R r ((C', R'), Is'); intros.
  assert (C'=C).
    mcinv evalp.
  mcrewrite C'; mcrewriteIn C' evalp; mcclear C'.

  mcinv evalp;

  mcclear5 C0 R0 r8 i0 f0; mcclear3 ns0 Is0 Is'0;
  mcinvClear wf_progstate; mcinvClear wf_iseq;
  apply wf_ps with G; auto;
  try mcrewrite R'; try mcrewrite Is'.
  
  mcinvClear wf_codeheap. generalize (H3 _ _ H8); intros.
  mcunfoldInSelf wf_codeval.
  assert (wf_iseq C G' Is').
    mcinvClear wf_regfile.
    generalize (H11 r); mcrewrite (G r); intros.
    repeat mcrewriteIn R' eq; mcrewriteIn R' evalp; mcclear R'.
    generalize (wf_wval_tcode_app_ctype _ _ _ _ _ _ H2 H6 H7 H8);
      unfold ctypeOf, projS1; intros.
    apply (wf_ciseq_app_wf_iseq _ _ _ _ _ H9 _ H4 H1).
  apply wf_iseq_subrftype with G'; auto; try mcrewrite Is.
Qed.


(* Complete theorems *)

Theorem progress 
  : forall P, wf_progstate P -> exists P', evalp P P'.
Proof.
  intros ((C, R), Is); elim Is; clear Is. 

  intro i; case i; intros.
  eapply progress_iadd; eauto.
  eapply progress_iaddi; eauto.
  eapply progress_isub; eauto.
  eapply progress_isubi; eauto.
  eapply progress_imov; eauto.
  eapply progress_imovi; eauto.
  eapply progress_imovf; eauto.
  eapply progress_ibgti; eauto.
  eapply progress_itappi; eauto.
  intros; eapply progress_ijd; eauto.
  intros; eapply progress_ijmp; eauto.
Qed.

Theorem preservation 
  : forall P P', wf_progstate P -> evalp P P' -> wf_progstate P'.
Proof.
  intros ((C, R), Is); elim Is; clear Is.

  intro i; case i; intros.
  generalize preserv_iadd; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_iaddi; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_isub; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_isubi; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_imov; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_imovi; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_imovf; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_ibgti; unfold preservStatement; intros HH; eapply HH; eauto.
  generalize preserv_itappi; unfold preservStatement; intros HH; eapply HH; eauto.
  intros; eapply preserv_ijd; eauto.
  intros; eapply preserv_ijmp; eauto.
Qed.









